module overlap_m

  implicit none

contains

  subroutine overlap(flow, unit_edgelist, nlon, nlat, periodic, dist_lim, &
       hshp, j, k, delta)

    ! Finds edges between flow(j - delta) and flow(j), corresponding
    ! to dates k - delta and k. Calls write_overlap for these
    ! edges. Updates flow(j - delta)%list_vis%delta_out,
    ! flow(j)%list_vis%delta_in and flow(j - delta + 1:j -
    ! 1)%number_eddies.

    ! Libraries:
    use contour_531, only: polyline
    use gpc_f, only: gpc_polygon_clip_f, GPC_INT, polygon
    use nr_util, only: twopi

    use candidate_overlap_m, only: candidate_overlap
    use derived_types, only: snapshot, shp_triplet
    use spher_polygon_area_m, only: spher_polygon_area
    use spher_polyline_area_m, only: spher_polyline_area
    use weight_m, only: weight
    use write_overlap_m, only: write_overlap

    type(snapshot), intent(inout):: flow(:) ! (max_delta + 1)
    integer, intent(in):: unit_edgelist ! logical unit for edgelist file
    integer, intent(in):: nlon, nlat
    logical, intent(in):: periodic ! grid is periodic in longitude

    integer, intent(in):: dist_lim
    ! We look for an overlapping eddy at dist_lim (in grid points) of
    ! the extremum of a given eddy.

    ! Triplet of shapefiles to which any interpolated eddy will be written:
    TYPE(shp_triplet), intent(in):: hshp

    integer, intent(in):: j
    ! position in time window, between 2 and max_delta + 1

    integer, intent(in):: k ! date index
    integer, intent(in):: delta ! between 1 and max_delta

    ! Local:

    integer i1 ! eddy index of predecessor
    integer i2 ! eddy index of successor
    integer l, n_select, m
    type(polyline) polyline_1, polyline_2
    type(polygon) res_pol

    integer, allocatable:: selection(:)
    ! identifying numbers of a selection of extrema

    ! Window around each extremum:
    integer llc(2) ! indices in global grid of lower left corner
    integer urc(2) ! indices in global grid of upper right corner

    !-----------------------------------------------------------------------

    loop_i1: do i1 = 1, flow(j - delta)%number_vis_extr
       test_valid: if (flow(j - delta)%list_vis(i1)%valid) then
          ! Define the geographical window around each eddy extremum:

          llc = flow(j - delta)%ind_extr(:, i1) - dist_lim
          urc = flow(j - delta)%ind_extr(:, i1) + dist_lim

          llc(2) = max(llc(2), 1)
          urc(2) = min(urc(2), nlat)

          if (.not. periodic) then
             llc(1) = max(llc(1), 1)
             urc(1) = min(urc(1), nlon)
          end if

          ! Pre-select potential successors:
          selection = candidate_overlap(flow(j)%extr_map(llc(1):urc(1), &
               llc(2):urc(2)), flow(j)%list_vis, &
               flow(j - delta)%list_vis(i1)%cyclone, &
               flow(j - delta)%list_vis(i1)%delta_out, delta)
          
          n_select = size(selection)
          if (n_select /= 0) polyline_1 &
               = merge(flow(j - delta)%list_vis(i1)%speed_cont%polyline, &
               flow(j - delta)%list_vis(i1)%out_cont%polyline, &
               flow(j - delta)%list_vis(i1)%speed_cont%n_points /= 0)

          DO l = 1, n_select
             i2 = selection(l)
             ! Assertion: {flow(j - delta)%list_vis(i1)%delta_out >=
             ! delta .or. flow(j)%list_vis(i2)%delta_in >= delta}
             polyline_2 = merge(flow(j)%list_vis(i2)%speed_cont%polyline, &
                  flow(j)%list_vis(i2)%out_cont%polyline, &
                  flow(j)%list_vis(i2)%speed_cont%n_points /= 0)

             ! Shift the longitudes of polyline_2 to values close to the
             ! longitude of extremum i1:
             m = floor((flow(j - delta)%list_vis(i1)%coord_extr(1) &
                  - flow(j)%list_vis(i2)%coord_extr(1)) / twopi + 0.5)
             polyline_2%points(1, :) = polyline_2%points(1, :) + m * twopi

             call gpc_polygon_clip_f(GPC_INT, &
                  polygon(nparts = 1, part = [polyline_1], hole = [.false.]), &
                  polygon(nparts = 1, part = [polyline_2], hole = [.false.]), &
                  res_pol)

             if (res_pol%nparts /= 0) then
                ! polyline_1 and polyline_2 overlap
                if (spher_polygon_area(res_pol) >= 0.25 &
                     * min(spher_polyline_area(polyline_1), &
                     spher_polyline_area(polyline_2))) then
                   ! Count one more fictitious eddy for all dates
                   ! between j - delta + 1 and j - 1:
                   flow(j - delta + 1:j - 1)%number_eddies &
                        = flow(j - delta + 1:j - 1)%number_eddies + 1

                   call write_overlap(hshp, unit_edgelist, k, delta, i1, i2, &
                        m, flow(j - delta + 1:j - 1)%number_eddies, &
                        flow(j - delta)%list_vis(i1)%cyclone, &
                        flow(j - delta)%list_vis(i1)%coord_extr, &
                        flow(j - delta)%list_vis(i1)%ssh_extr, &
                        flow(j)%list_vis(i2)%coord_extr, &
                        flow(j)%list_vis(i2)%ssh_extr, &
                        weight(flow(j - delta)%list_vis(i1), &
                        flow(j)%list_vis(i2)))
                   flow(j - delta)%list_vis(i1)%delta_out &
                        = min(flow(j - delta)%list_vis(i1)%delta_out, delta)
                   flow(j)%list_vis(i2)%delta_in &
                        = min(flow(j)%list_vis(i2)%delta_in, delta)
                end if
             end if
          end DO
       end if test_valid
    end do loop_i1

  end subroutine overlap
  
end module overlap_m
